/*
 * Run a program using the supplied arguments.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <port.h>
#include <process.h>



/* Transfer size for communication. */

static const offset_t TO_TRANSFER = 1024;

/* Transfer data between streams. */

static void transfer(FILE *input, FILE *output)
{
    char buffer[TO_TRANSFER];
    size_t nread;

    while ((nread = fread(buffer, sizeof(char), TO_TRANSFER, input)))
    {
        if (!fwrite(buffer, sizeof(char), nread, output))
            break;
    }
}



/* List objects in the filesystem image. */

int main(int argc, char *argv[])
{
    FILE *input, *output, *error;
    pid_t pid;
    port_t port;
    event_source_t source;
    event_values_t result;

    if (argc < 2)
    {
        printf("Usage: test_popenv <program> <argument>...\n");
        return 1;
    }

    printf("Running: %s\n", argv[1]);

    pid = popenv(argc - 1, (const char **) argv + 1, &input, &output, &error);

    if (pid == -1)
    {
        printf("Error: %d\n", errno);
        return 1;
    }

    printf("Process: %d\n", pid);

    /* Monitor the streams and the process. */

    port = port_create();

    port_subscribe(port, port_process_source(pid), PORT_PROCESS_ALL);

    if (stdin != NULL)
        port_subscribe(port, port_file_source(fileno(stdin)), PORT_INPUT | PORT_CLOSE);

    port_subscribe(port, port_file_source(fileno(output)), PORT_INPUT | PORT_CLOSE);

    if (error != NULL)
        port_subscribe(port, port_file_source(fileno(error)), PORT_INPUT | PORT_CLOSE);

    /* Wait for events on the streams and the process. */

    while (1)
    {
        if (port_wait(port, &source))
        {
            printf("Error waiting for notifications.\n");
            return 1;
        }

        /* Handle input to this process. */

        if ((stdin != NULL) && (source == port_file_source(fileno(stdin))))
            transfer(stdin, input);

        /* Handle output from the process. */

        else if (source == port_file_source(fileno(output)))
            transfer(output, stdout);
        else if ((error != NULL) && (source == port_file_source(fileno(error))))
            transfer(error, stderr);

        /* Handle process termination. */

        else if (source == port_process_source(pid))
        {
            transfer(output, stdout);

            if (error != NULL)
                transfer(error, stderr);

            result = port_source_event_values(source);
            break;
        }
    }

    fclose(input);
    fclose(output);

    if (error != NULL)
        fclose(error);

    printf("Result: %ld\n", result.val);
    return 0;
}

/* vim: tabstop=4 expandtab shiftwidth=4
*/
